#!/usr/bin/env python

"""
Farm out processing to multiple processes via zmq.
"""

import re, time, threading
import multiprocessing as mp

import zmq

from zerogroup import ZeroGroup

data = "demo foo bar not baz quux"
s_list = ['demo', 'foo', 'bar', 'quux']

class MyProcess(mp.Process):
    def find_str(self, s, data):
        res = re.search(s, data, re.MULTILINE)
        if res is not None:
            return res.start()

    def run(self):
        g = ZeroGroup()
        g.add(zmq.REP, "tcp://localhost:5555", False,
              {}, 'data', True)
        g.add(zmq.SUB, "tcp://localhost:5556", False,
              {zmq.SUBSCRIBE: ''}, 'ctrl', True)
        g.create()

        # Note that the control socket stream isn't flushed when a quit signal
        # is received because doing so causes errors:
        def ctrl_handler(msg):
            if msg[0] == 'quit':
                g.flush('data')
                g.ioloop.stop()
        g.on_recv('ctrl', ctrl_handler)

        def data_handler(msg):
            s, d = msg
            g.send_multipart('data', [s, str(self.find_str(s, d))])
        g.on_recv('data', data_handler)
        g.start_loop()


if __name__ == '__main__':

    # This must come before the sockets are added to the main parent
    # process:
    p_list = [MyProcess() for i in xrange(2)]
    map(lambda p: p.start(), p_list)

    g = ZeroGroup()
    g.add(zmq.DEALER, "tcp://*:5555", True, {}, 'data')
    g.add(zmq.PUB, "tcp://*:5556", True, {}, 'ctrl')
    g.create()

    # Send data to workers for processing:
    results = []
    for s in s_list:
        g.send_multipart('data', ['', s, data])
    
    # Wait for all results to arrive:
    while len(results) < len(s_list):
        if g.poll('data', 10):
            _, s, d = g.recv_multipart('data')
            results.append((s, d))

    # Send quit signals until all workers stop running:
    while p_list:
        g.send('ctrl', 'quit')
        for p in p_list:
            if not p.is_alive():
                p_list.remove(p)

    # Display results:
    print results



#!/usr/bin/env python

"""
Show how to prevent deadlocking because of traffic asymmetry
by sending null messages for a while after signalling a quit.
"""

import itertools, logging, time
import threading as th
import multiprocessing as mp
import zmq

import zerogroup

PORT_DATA = 6000
PORT_CTRL = 6001

def a_func():
    logger = logging.getLogger('a')

    g = zerogroup.ZeroGroup()
    g.add(zmq.ROUTER, 'tcp://*:%i' % PORT_DATA, True, {}, 'data')
    g.add(zmq.DEALER, 'tcp://localhost:%i' % PORT_CTRL, False, 
          {zmq.IDENTITY: 'a'}, 'ctrl', True)
    g.create()

    def handler(msg):
        if msg[0] == 'quit':
            g.flush('ctrl')
            g.stop_loop()

            handler.running = False
            logger.info('recv ctrl quit')
    handler.running = True
    g.on_recv('ctrl', handler)
    g.start_loop(True)

    c = itertools.count()
    while True:
        data = str(c.next())
        g.send_multipart('data', ['b', data])
        g.send('data', data)
        logger.info('sent data %s' % data)

        _, data = g.recv_multipart('data')
        logger.info('recv data %s' % data)

        # Send null data when exiting to prevent destination node from
        # hanging on recv:
        if not handler.running:
            for i in xrange(5):
                g.send_multipart('data', ['b', 'NULL'])
            break
        
def b_func():
    logger = logging.getLogger('b')

    g = zerogroup.ZeroGroup()
    g.add(zmq.DEALER, 'tcp://localhost:%i' % PORT_DATA, False,
          {zmq.IDENTITY: 'b'}, 'data')
    g.add(zmq.DEALER, 'tcp://localhost:%i' % PORT_CTRL, False,
          {zmq.IDENTITY: 'b'}, 'ctrl', True)
    g.create()

    def handler(msg):
        if msg[0] == 'quit':
            g.flush('ctrl')
            g.stop_loop()
            handler.running = False
            logger.info('recv ctrl quit')
    handler.running = True
    g.on_recv('ctrl', handler)
    g.start_loop(True)

    c = itertools.count()
    while True:
        data = str(c.next())
        g.send('data', data)
        logger.info('sent data %s' % data)
        time.sleep(0.01)

        data = g.recv('data')
        logger.info('recv data %s' % data)

        # Send null data when exiting to prevent destination node from
        # hanging on recv:        
        if not handler.running:
            for i in xrange(5):
                g.send('data', 'NULL')
            break
        
if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(name)s %(levelname)s %(message)s')

    ctx = zmq.Context()
    sock = ctx.socket(zmq.ROUTER)
    sock.bind('tcp://*:%i' % PORT_CTRL)

    a = mp.Process(target=a_func)
    a.start()

    b = mp.Process(target=b_func)
    b.start()

    time.sleep(2)

    for i in ['a', 'b']:
        sock.send_multipart([i, 'quit'])




#!/usr/bin/env python

import sys, os
from glob import glob

try:
    import setuptools
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()

from setuptools import find_packages
from setuptools import setup

NAME =               'zerogroup'
VERSION =            '0.1.0'
AUTHOR =             'Lev Givon'
AUTHOR_EMAIL =       'lev@columbia.edu'
URL =                'https://github.com/lebedov/zerogroup'
DESCRIPTION =        'ZeroMQ socket group wrapper'
LONG_DESCRIPTION =   DESCRIPTION
DOWNLOAD_URL =       URL
LICENSE =            'BSD'
CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Topic :: Software Development']
PACKAGES = find_packages()

if __name__ == "__main__":
    if os.path.exists('MANIFEST'):
        os.remove('MANIFEST')

    setup(
        name = NAME,
        version = VERSION,
        author = AUTHOR,
        author_email = AUTHOR_EMAIL,
        license = LICENSE,
        classifiers = CLASSIFIERS,
        description = DESCRIPTION,
        long_description = LONG_DESCRIPTION,
        url = URL,
        packages = PACKAGES,
        install_requires = ['pyzmq'],
        extras_require = {'msgpack': ['msgpack-python>=0.4.0']}
        )



#!/usr/bin/env python

"""
ZeroMQ socket group wrapper.
"""

from .version import __version__

import re
import threading as th

# Non-standard serialization support not mandatory:
try:
    import msgpack
except:
    pass

import zmq
from zmq.eventloop.ioloop import IOLoop
from zmq.eventloop.zmqstream import ZMQStream

def istransport(t):
    """
    Return True of `t` is a valid ZeroMQ transport.
    """

    try:
        status = bool(re.match('^(inproc|ipc|tcp|pgm|epgm)\://.*', t))
    except:
        return False
    else:
        return status

class ZeroGroup(object):
    """
    ZeroMQ socket group wrapper.

    Encapsulates a group of ZeroMQ sockets (and associated streams)
    sharing the same context and (for those sockets with streams) event loop.

    Parameters
    ----------
    context : zmq.Context
        ZeroMQ context to use. If no context is specified, a
        new one is created.
    ioloop : zmq.eventloop.ioloop.ZMQIOLoop
        IOLoop instance to use for event streams. If no instance is specified,
        a new one is created.

    Example
    -------
    >>> import zmq
    >>> g = SocketGroup()
    >>> g.add(zmq.DEALER, 'ipc://data', True, {zmq.IDENTITY: 'xyz'}, 'data')
    >>> g.add(zmq.DEALER, 'ipc://ctrl', True, {zmq.IDENTITY: 'xyz'}, 'ctrl')
    >>> g.create()
    >>> msg = g.recv('data')
    >>> g.send('ctrl', 'foo')
    """

    def __init__(self, context=None, ioloop=None):
        if context is None:
            self.context = zmq.Context()
        elif isinstance(context, zmq.Context):
            self.context = context
        else:
            raise ValueError('invalid context')
        if ioloop is None:
            self.ioloop = IOLoop.instance()
        elif isinstance(ioloop, IOLoop):
            self.ioloop = ioloop
        else:
            raise ValueError('invalid ioloop')

        self._created = False

        # Specifications of sockets to create:
        self.specs = {}

        # Maps addresses to sockets:
        self.sock = {}

        # Maps addresses to streams:
        self.stream = {}

        # Maps string aliases to addresses:
        self.alias = {}

    def _getaddr(self, n):
        """
        Convert alias to address.
        """

        if n in self.alias:
            return self.alias[n]
        elif n in self.sock:
            return n
        else:
            raise ValueError('unknown alias/address')

    def add(self, socket_type, addr, binding, 
            opts={}, alias=None, stream=False):
        """
        Add a new socket to the group.
        """

        if not istransport(addr):
            raise ValueError('invalid address')
        if alias and istransport(alias):
            raise ValueError('can only alias non-address to address')
        if not isinstance(opts, dict):
            raise ValueError('invalid options')
        self.specs[addr] = {'socket_type': socket_type,
                            'binding': binding,
                            'stream': stream,
                            'alias': alias,
                            'opts': opts}

    def create(self):
        """
        Create all specified sockets and streams.
        """

        if self._created:
            return
        for addr in self.specs:
            spec = self.specs[addr]
            sock = self.context.socket(spec['socket_type'])

            # Set socket options before binding:
            for opt, val in spec['opts'].iteritems():
                sock.setsockopt(opt, val)
            if spec['binding']:
                sock.bind(addr)
            else:
                sock.connect(addr)

            if spec['stream']:
                stream = ZMQStream(sock, self.ioloop)
                self.stream[addr] = stream
            else:
                stream = None

            self.sock[addr] = sock
            alias = spec['alias']
            if alias:
                self.alias[alias] = addr

                # Insert alias into self.stream so that it can be used to lookup 
                # a stream directly via self.stream[alias]:
                self.stream[alias] = addr
        self._created = True

    def poll(self, n, timeout=None, flags=1):
        """
        Poll socket for events.
        """

        return self.sock[self._getaddr(n)].poll(timeout, flags)

    def poll_in(self, n, timeout=None):
        """
        Poll socket for incoming events.
        """

        return self.sock[self._getaddr(n)].poll(timeout, zmq.POLLIN)

    def poll_out(self, n, timeout=None):
        """
        Poll socket for outgoing events.
        """

        return self.sock[self._getaddr(n)].poll(timeout, zmq.POLLOUT)

    def poll_all(self, n, timeout=None):
        """
        Poll socket for all events.
        """

        return self.sock[self._getaddr(n)].poll(timeout, zmq.POLLIN|zmq.POLLOUT)

    def send(self, n, data, flags=0, copy=True, track=False):
        """
        Send a message.
        """

        return self.sock[self._getaddr(n)].send(data, flags, copy, track)

    def send_multipart(self, n, msg_parts, flags=0, copy=True, track=False):
        """
        Send a multipart message.
        """

        return self.sock[self._getaddr(n)].send_multipart(msg_parts, flags, copy, track)

    def send_pyobj(self, n, obj, flags=0, protocol=-1):
        """
        Send a Python object as a message using pickle to serialize.
        """

        return self.sock[self._getaddr(n)].send_pyobj(obj, flags, protocol)

    def send_msgpack(self, n, obj, flags=0):
        """
        Send a Python object as a message using msgpack to serialize.
        """

        msg = msgpack.dumps(obj)
        return self.send(n, msg, flags)

    def recv(self, n, flags=0, copy=True, track=False):
        """
        Receive a message.
        """
        return self.sock[self._getaddr(n)].recv(flags, copy, track)

    def recv_multipart(self, n, flags=0, copy=True, track=False):
        """
        Receive a multipart message.
        """
        return self.sock[self._getaddr(n)].recv_multipart(flags, copy, track)

    def recv_pyobj(self, n, flags=0):
        """
        Receive a Python object serialized with pickle.
        """

        return self.sock[self._getaddr(n)].recv_pyobj(flags)

    def recv_msgpack(self, n, flags=0):
        """
        Receive a Python object serialized with msgpack.
        """

        s = self.sock[self._getaddr(n)].recv(flags)
        return msgpack.loads(s)

    def flush(self, n, flags=3, limit=None):
        """
        Flush stream.
        """

        return self.stream[self._getaddr(n)].flush(flags, limit)

    def on_recv(self, n, callback, copy=True):
        """
        Set stream receive handler.
        """

        self.stream[self._getaddr(n)].on_recv(callback, copy)

    def on_send(self, n, callback):
        """
        Set stream send handler.
        """

        self.stream[self._getaddr(n)].on_send(callback)

    def start_loop(self, threaded=False):
        """
        Start event loop.

        Parameters
        ----------
        threaded : bool
            If True, start the loop in a new thread.
        """

        if threaded:
            th.Thread(target=self.ioloop.start).start()
        else:
            self.ioloop.start()

    def stop_loop(self, flush=False, stop_on_recv=False, stop_on_send=False):
        """
        Stop event loop.

        Parameters
        ----------
        flush : bool
            Flush all streams before stopping the loop.
        stop_on_recv, stop_on_send : bool
            Run each stream's stop_on_recv() or stop_on_send() methods
            before stopping the loop.
        """

        for stream in self.stream.values():
            if flush:
                stream.flush()
            if stop_on_recv:
                stream.stop_on_recv()
            if stop_on_send:
                stream.stop_on_send()
        self.ioloop.stop()

    def __getitem__(self, n):
        return self.sock[self._getaddr(n)]

    def __iter__(self):
        for addr in self.sock.keys():
            yield addr
